// DataTypeUtils.java
//
// (c) 1999-2001 PAL Development Core Team
//
// This package may be distributed under the
// terms of the Lesser GNU General Public License (LGPL)

package pal.datatype;

import java.io.Serializable;

/**
 * helper class for sequence data types.
 *
 * @version $Id: DataTypeUtils.java,v 1.6 2001/11/20 07:26:43 matt Exp $
 *
 * @author Korbinian Strimmer
 * @author Alexei Drummond
 */
public class DataTypeUtils implements Serializable
{
	//
	// Public stuff
	//

	/**
	 * create object according to this code
	 *
	 * @param typeID selected data type
	 *
	 * @return DataType object
	 */
	public static DataType getInstance(int typeID)
	{
		if (typeID == DataType.NUCLEOTIDES) {
			return new Nucleotides();
		}
		else if (typeID == DataType.AMINOACIDS) {
			return new AminoAcids();
		} else if (typeID == DataType.TWOSTATES) {
			return new TwoStates();
		} else if (typeID == DataType.IUPACNUCLEOTIDES) {
			return new IUPACNucleotides();
		} else if (typeID == DataType.CODONS) {
			return new Codons();
		} else if (typeID == DataType.GENERALIZEDCODONS) {
			return GeneralizedCodons.DEFAULT_INSTANCE;
		} else {
			throw new IllegalArgumentException("typeID " + typeID + " is not recognised.");
		}
	}

	/**
	 * Return states of a sequence.
	 */
	public final static int[] getSequenceStates(DataType d, String sequence) {

		int[] states = new int[sequence.length()];

		for (int i = 0; i < sequence.length(); i++) {
			states[i] = d.getState(sequence.charAt(i));
		}
		return states;
	}

	/**
	 * returns true if the character represents a gap in the sequence.
	 */
	public final static boolean isGap(DataType d, char c) {
		return d.getState(c) == d.getNumStates();
	}

	/** For converting a sequence to an array of bytes where each byte represents the
	 *		state of the corresponding character in sequence
	 */
	public final static byte[] getByteStates(String sequence , DataType dt) {
		return getByteStates(sequence.toCharArray(),dt);
	}

	/** For converting a sequence to an array of bytes where each byte represents the
	 *		state of the corresponding character in sequence
	 */
	public final static byte[] getByteStates(char[] sequence , DataType dt) {
		byte[] bs = new byte[sequence.length];
		for(int i = 0 ; i < bs.length ; i++) {
			bs[i] = (byte)dt.getState(sequence[i]);
		}
		return bs;
	}
	/** For converting an array of sequence to arrays of ints where each int represents the
	 *		state of the corresponding character in sequence
	 */
	public final static int[][] getStates(char[][] sequences, DataType dt) {
		int[][] statesSeqs = new int[sequences.length][];
		for(int i = 0  ; i < statesSeqs.length ; i++) {
			statesSeqs[i] = getStates(sequences[i],dt);
		}
		return statesSeqs;
	}
	/** For converting a sequence to an array of ints where each int represents the
	 *		state of the corresponding character in sequence
	 */
	public final static int[] getStates(char[] sequence, DataType dt) {
		int[] states= new int[sequence.length];
		for(int i = 0 ; i < states.length ; i++) {
			states[i] = dt.getState(sequence[i]);
		}
		return states;
	}

	/** For converting an array of sequence to arrays of ints where each int represents the
	 *		state of the corresponding character in sequence
	 *    @param unknownState ensures that the state representation is set to this value (like -1)
	 */
	public final static int[][] getStates(char[][] sequences, int unknownState, DataType dt) {
		int[][] statesSeqs = new int[sequences.length][];
		for(int i = 0  ; i < statesSeqs.length ; i++) {
			statesSeqs[i] = getStates(sequences[i], unknownState,dt);
		}
		return statesSeqs;
	}
	/** For converting a sequence to an array of ints where each int represents the
	 *		state of the corresponding character in sequence
	 */
	public final static int[] getStates(char[] sequence, int unknownState, DataType dt) {
		int[] states= new int[sequence.length];
		for(int i = 0 ; i < states.length ; i++) {

			states[i] = dt.getState(sequence[i]);
			if(dt.isUnknownState(states[i])) {
				states[i] = unknownState;
			}
		}
		return states;
	}

	/** For converting an array of sequence to arrays of ints where each int represents the
	 *		state of the corresponding character in sequence
	 *    @param unknownChar The character uses for unknown states
	 */
	public final static char[][] getChars(int[][] sequences, char unknownChar, DataType dt) {
		char[][] charSeqs = new char[sequences.length][];
		for(int i = 0  ; i < charSeqs.length ; i++) {
			charSeqs[i] = getChars(sequences[i], unknownChar,dt);
		}
		return charSeqs;
	}
	/** For converting a sequence of ints representing states to an array of chars
	 *  @param unknownChar The character uses for unknown states
	 */
	public final static char[] getChars(int[] sequence, char unknownChar, DataType dt) {
		char[] chars= new char[sequence.length];
		for(int i = 0 ; i < chars.length ; i++) {
			if(dt.isUnknownState(sequence[i])) {
				chars[i] = unknownChar;
			} else {
				chars[i] = dt.getChar(sequence[i]);
			}
		}
		return chars;
	}

}
