// AlignmentDistanceMatrix.java
//
// (c) 1999-2001 PAL Development Core Team
//
// This package may be distributed under the
// terms of the Lesser GNU General Public License (LGPL)

// Known bugs and limitations:
// - computational complexity of order O(numSeqs^2)


package pal.distance;

import pal.alignment.*;
import pal.substmodel.*;
import java.io.*;


/**
 * compute distance matrix (observed and ML) from alignment (SitePattern)
 *
 * @version $Id: AlignmentDistanceMatrix.java,v 1.7 2001/09/10 05:26:02 matt Exp $
 *
 * @author Korbinian Strimmer
 * @author Alexei Drummond
 */
public class AlignmentDistanceMatrix extends DistanceMatrix implements Serializable {
	//
	// Public stuff
	//

	/**
	 * compute observed distances
	 *
	 * @param sp site pattern
	 */
	public AlignmentDistanceMatrix(SitePattern sp)
	{
    this(sp,null,true);
	}
  /**
	 * compute observed distances
   *
	 * @param sp site pattern
   * @param immediateCompute - signifies whether to calculate distances from within constructor
   *	(if no should call recompute() at some point!)
	 *
   */
	public AlignmentDistanceMatrix(SitePattern sp, boolean immediateCompute)
	{
    this(sp,null,immediateCompute);
	}
	/**
	 * compute maximum-likelihood distances
	 *
	 * @param sp site pattern
	 * @param m  evolutionary model
	 */
	public AlignmentDistanceMatrix(SitePattern sp, SubstitutionModel m) {
   	this(sp,m,true);
  }
	/**
	 * compute maximum-likelihood distances
	 *
	 * @param sp site pattern
	 * @param m  evolutionary model
   * @param immediateCompute - signifies whether to calculate distances from within constructor
   *	(if no should call recompute() at some point!)
	 */
	protected AlignmentDistanceMatrix(SitePattern sp, SubstitutionModel m, boolean immediateCompute)
	{
		numSeqs = sp.getSequenceCount();
		idGroup = sp;
		distance = new double[numSeqs][numSeqs];

		if(m!=null) {
    	pwd = new PairwiseDistance(sp, m);
    } else {
    	pwd = new PairwiseDistance(sp);
    }
    if(immediateCompute) {
			computeDistances();
    }
	}

	/**
	 * recompute maximum-likelihood distances under new model
	 *
	 * @param m  evolutionary model
	 */
	public void recompute(SubstitutionModel m)
	{
		pwd.updateModel(m);

		computeDistances();
	}

	/**
	 * recompute maximum-likelihood distances under new site pattern
	 *
	 * @param sp site pattern
	 */
	public void recompute(SitePattern sp)
	{
		pwd.updateSitePattern(sp);

		computeDistances();
	}


	//
	// Private stuff
	//

	private PairwiseDistance pwd;

	private void computeDistances()
	{
		for (int i = 0; i < numSeqs; i++)
		{
			distance[i][i] = 0;
			for (int j = i + 1; j < numSeqs; j++)
			{
      	progress_  = 2*(i*numSeqs+j)/((double)(numSeqs*numSeqs));
				distance[i][j] = pwd.getDistance(i, j);
				distance[j][i] = distance[i][j];
			}
		}
    progress_ = -1;		
	}

  /** Returns -1 if not doing anything, or a number between 0 and 1 */
  public final double getProgress() {
  	return progress_;
  }
  private double progress_ = -1;

}
