// ParameterizedDouble.java
//
// (c) 1999-2001 PAL Development Core Team
//
// This package may be distributed under the
// terms of the Lesser GNU General Public License (LGPL)


package pal.misc;

import java.io.*;


/**
 * interface for a double that might be used as a parameter
 *
 * @version $Id: ParameterizedDouble.java,v 1.4 2001/08/04 20:02:31 alexi Exp $
 *
 * @author Matthew Goode
 */

import java.util.*;

public class ParameterizedDouble implements Parameterized, java.io.Serializable{
	private double defaultValue_;
	private double currentValue_;
	private double minimumValue_;
	private double maximumValue_;
	private double se_;

	private Vector listeners_ = new Vector();

	/** The default value is also the initial value.
	*/
	public ParameterizedDouble(double defaultValue, double minimumValue, double maximumValue) {
		this.currentValue_ = defaultValue;
		this.defaultValue_ = defaultValue;
		this.minimumValue_ = minimumValue;
		this.maximumValue_ = maximumValue;
	}

	/** Set the current value of this double */
	public final void setValue(double value) {
		this.currentValue_ = value;
		fireValueChangedEvent();
	}

	/** Get the current value of this double */
	public final double getValue() {
		return currentValue_;
	}

	public final void setMaximumValue(double value) {
		this.maximumValue_ =value;
	}
	public final void setMinimumValue(double value) {
		this.minimumValue_ =value;
	}

	private final void fireValueChangedEvent() {
		ParameterEvent pe = new ParameterEvent(this);
		for(Enumeration e = listeners_.elements() ; e.hasMoreElements() ; ) {
			((ExternalParameterListener)e.nextElement()).parameterChanged(pe);
		}
	}

	public final void addExternalParameterListener(ExternalParameterListener l) {
		if(!listeners_.contains(l)) {
			listeners_.addElement(l);
		}
	}
	public final void removeExternalParameterListener(ExternalParameterListener l) {
		listeners_.removeElement(l);
	}
	public final double getLowerLimit() {
		return minimumValue_;
	}

	public final double getUpperLimit() {
		return maximumValue_;
	}

	public final double getDefaultValue() {
		return defaultValue_;
	}

	public final double getSE() {
		return se_;
	}

	public final void setSE(double value) {
		se_ = value;
	}

	/**
	 *
	 * @return 1
	 */
	public int getNumParameters() {
		return 1;
	}

	/**
	 * set model parameter
	 *
	 * @param param  parameter value
	 * @param n  parameter number (ignored)
	 */
	public final void setParameter(double param, int n) {
		setValue(param);
	}

	/**
	 * get model parameter
	 *
	 * @param n  parameter number
	 *
	 * @return parameter value
	 */
	public double getParameter(int n) {
		return currentValue_;
	}


	/**
	 * set standard errors for model parameter
	 *
	 * @param paramSE  standard error of parameter value
	 * @param n parameter number
	 */
	public void setParameterSE(double paramSE, int n) {
		this.se_ = paramSE;
	}

	
	/**
	 * get lower parameter limit
	 *
	 * @param n parameter number
	 *
	 * @return lower bound
	 */
	public double getLowerLimit(int n) {
   	return minimumValue_;
	}

	/**
	 * get upper parameter limit
	 *
	 * @param n parameter number
	 *
	 * @return upper bound
	 */
	public double getUpperLimit(int n) {
		return maximumValue_;
	}


	/**
	 * get default value of parameter
	 *
	 * @param n parameter number
	 *
	 * @return default value
	 */
	public double getDefaultValue(int n) {
		return defaultValue_;
	}
	public String toString() {
		return "Parameterized Double (current,default,min,max):"
				+currentValue_+", "
				+defaultValue_+", "
				+minimumValue_+", "
				+maximumValue_+", ";
	}
}
