// SimpleSubstitutionModel.java
//
// (c) 1999-2001 PAL Development Core Team
//
// This package may be distributed under the
// terms of the Lesser GNU General Public License (LGPL)


package pal.substmodel;

import pal.misc.*;
import pal.math.*;
import pal.datatype.*;

import java.io.*;


/**
 * <b>model of sequence substitution (rate matrix + rate variation)</b>.
 * provides a convenient interface for the computation of transition probabilities
 *
 * @version $Id: SimpleSubstitutionModel.java,v 1.7 2001/10/10 07:50:02 matt Exp $
 *
 * @author Korbinian Strimmer
 * @author Alexei Drummond
 */
public class SimpleSubstitutionModel implements SubstitutionModel {

	/** rate matrix */
	private RateMatrix rateMatrix;

	/** rate distribution */
	private RateDistribution rateDistribution;

	/** dimension of rate matrix */
	private int dimension;

	/** number of categories in discrete rate distribution */
	private int numRates;

	private double[][][] probs;
	private int numRmatParams, numRdistParams, numParams;

	//
	// Serialization code
	//
	private static final long serialVersionUID= -4264547361778477341L;
	//serialver -classpath ./classes pal.substmodel.SubstitutionModel

	private void writeObject(java.io.ObjectOutputStream out) throws java.io.IOException {
		out.writeByte(1); //Version number
		out.writeObject(rateMatrix);
		out.writeObject(rateDistribution);
		out.writeObject(probs);
	}

	private void readObject(java.io.ObjectInputStream in) throws java.io.IOException, ClassNotFoundException{
		byte version = in.readByte();
		switch(version) {
			default : {
				rateMatrix = (RateMatrix)in.readObject();
				rateDistribution = (RateDistribution)in.readObject();
				probs = (double[][][])in.readObject();
				dimension = rateMatrix.getDimension();
				numRates = rateDistribution.numRates;
				numRmatParams = rateMatrix.getNumParameters();
				numRdistParams = rateDistribution.getNumParameters();
				numParams = numRmatParams + numRdistParams;
				 break;
			}
		}
	}

	/**
	 * constructor 1 (uniform rate distribution)
	 *
	 * @param rmat rate matrix
	 */
	public SimpleSubstitutionModel(RateMatrix rmat)
	{
		this(rmat, new UniformRate());
	}

	/**
	 * constructor 2 (gamma rate distribution)
	 *
	 * @param rmat rate matrix
	 * @param n number of rate categories
	 * @param a shape parameter
	 */
	public SimpleSubstitutionModel(RateMatrix rmat, int n, double a)
	{
		this(rmat, new GammaRates(n, a));
	}

	/**
	 * constructor 3 (arbitrary rate distribution)
	 *
	 * @param rmat rate matrix
	 * @param rdist rate distribution
	 */
	public SimpleSubstitutionModel(RateMatrix rmat, RateDistribution rdist)
	{
		dimension = rmat.getDimension();
		numRates = rdist.numRates;

		rateMatrix = rmat;
		rateDistribution = rdist;

		probs = new double[numRates][dimension][dimension];

		numRmatParams = rmat.getNumParameters();
		numRdistParams = rdist.getNumParameters();
		numParams = numRmatParams + numRdistParams;
	}

	public DataType getDataType() {
		return rateMatrix.getDataType();
	}

	/**
	 * Return a copy of this substitution model.
	 */
	public SimpleSubstitutionModel(SubstitutionModel model) {

		dimension = model.getRateMatrix().getDimension();
		numRates = model.getRateDistribution().numRates;
		numRmatParams = model.getNumParameters();
		numRdistParams = model.getRateDistribution().getNumParameters();

		rateMatrix = (RateMatrix)model.getRateMatrix().clone();
		rateDistribution = (RateDistribution)model.getRateDistribution().clone();

		probs = new double[numRates][dimension][dimension];

		numParams = numRmatParams + numRdistParams;
	}

	// interface Report
	public void report(PrintWriter out)
	{
		rateMatrix.report(out);
		out.println();
		rateDistribution.report(out);
	}

	/**
	 * Return string representation of substitution model.
	 */
	public String toString() {
		StringWriter sw = new StringWriter();
		report(new PrintWriter(sw));
		return sw.toString();
	}

	// interface Parameterized

	public int getNumParameters()
	{
		return numParams;
	}

	public void setParameter(double param, int n)
	{
		if (n < numRmatParams)
		{
			rateMatrix.setParameter(param, n);
		}
		else
		{
			rateDistribution.setParameter(param, n-numRmatParams);
		}
	}

	public double getParameter(int n)
	{
		if (n < numRmatParams)
		{
			return rateMatrix.getParameter(n);
		}
		else
		{
			return rateDistribution.getParameter(n-numRmatParams);
		}
	}

	public void setParameterSE(double paramSE, int n)
	{
		if (n < numRmatParams)
		{
			rateMatrix.setParameterSE(paramSE, n);
		}
		else
		{
			rateDistribution.setParameterSE(paramSE, n-numRmatParams);
		}
	}

	public double getLowerLimit(int n)
	{
		if (n < numRmatParams)
		{
			return rateMatrix.getLowerLimit(n);
		}
		else
		{
			return rateDistribution.getLowerLimit(n-numRmatParams);
		}
	}

	public double getUpperLimit(int n)
	{
		if (n < numRmatParams)
		{
			return rateMatrix.getUpperLimit(n);
		}
		else
		{
			return rateDistribution.getUpperLimit(n-numRmatParams);
		}
	}

	public double getDefaultValue(int n)
	{
		if (n < numRmatParams)
		{
			return rateMatrix.getDefaultValue(n);
		}
		else
		{
			return rateDistribution.getDefaultValue(n-numRmatParams);
		}
	}

	public void setTime(double start, double end) {
		setDistance(Math.abs(start - end));
	}


	/**
	 * set distance and corresponding computation transition probabilities
	 *
	 * @param k distance
	 */
	public void setDistance(double k)
	{
		for (int r = 0; r < numRates; r++)
		{
			double kk = k*rateDistribution.rate[r];

			if (kk < BranchLimits.MINARC)
			{
				kk = BranchLimits.MINARC;
			}
			if (kk > BranchLimits.MAXARC)
			{
				kk = BranchLimits.MAXARC;
			}

			rateMatrix.setDistance(kk);
			rateMatrix.getTransitionProbabilities(probs[r]);
		}
	}

	/**
	 * get transition probability for the preselected model and
	 * the previously specified distance
	 *
	 * @param r rate category
	 * @param i start state
	 * @param j end state
	 *
	 * @return transition probability
	 */
	public double transProb(int r, int i, int j)
	{
		return probs[r][i][j];
	}

	public boolean isSimpleJukesCantor() {

		// if more then one rate the not a simple jukes cantor
		if (numRates > 1) return false;

		// NOT if frequencies are different
		double[] frequencies = rateMatrix.getEquilibriumFrequencies();
		double freq = frequencies[0];
		for (int i = 1; i < frequencies.length; i++) {
			if (frequencies[i] != freq) {
				return false;
			}
		}

		double[][] relRates = rateMatrix.getRelativeRates();
		// NOT if rates are different
		double rate = relRates[0][1];
		for (int i = 0; i < relRates.length; i++) {
			for (int j = 0; j < relRates[i].length; j++) {
				if ((i != j) && (relRates[i][j] != rate)) {
					return false;
				}
			}
		}

		return true;
	}

	/** A non shallow implementation of clone() */
	public Object clone() {
		return new SimpleSubstitutionModel(this);
	}

	public RateMatrix getRateMatrix() { return rateMatrix;}
	public RateDistribution getRateDistribution() { return rateDistribution;}
	public int getDimension() { return dimension; }

	/** Add a PalObjectListener to be notified of changes to the model.
	 *  Only the parametersChanged method will generally be called.
	 *  Currently passes request on to rateMatrix
	 */
	public void addPalObjectListener(PalObjectListener pol) { rateMatrix.addPalObjectListener(pol); }
	/**
	 * Currently passes request on to rateMatrix
	 */
	public void removePalObjectListener(PalObjectListener pol) { rateMatrix.removePalObjectListener(pol); }

}
