// DatedTipsClockTree.java
//
// (c) 1999-2001 PAL Development Core Team
//
// This package may be distributed under the
// terms of the Lesser GNU General Public License (LGPL)


package pal.tree;

import pal.misc.*;


/**
 * provides parameter interface to a clock tree with dated tips,
 * following A. Rambaut. 2000. Bioinformatics 16:395-399.
 * (parameters are the minimal node height differences
 * at each internal node and the evolutionary rate)
 *
 * @version $Id: DatedTipsClockTree.java,v 1.15 2001/11/20 07:35:44 matt Exp $
 *
 * @author Korbinian Strimmer
 * @author Alexei Drummond
 */
public class DatedTipsClockTree extends ParameterizedTree
	implements ExternalParameterListener,java.io.Serializable {

	//
	// Public stuff
	//

	public static final int PARAMETERIZE_WITH_DEFAULTS = 0;
	public static final int PARAMETERIZE_WITH_GIVEN = 1;
	public static final int NO_PARAMETRIZE = 2;


	/**
	 * Constructor without TimeOrderCharacterData.
	 * Dates are extracted from labels.
	 */
	public DatedTipsClockTree(Tree t) {
		this(t,  null, true);
	}

	/**
	 * take any tree and afford it with an interface
	 * suitable for a clock-like tree with dated tips (parameters
	 * are the minimal node height differences at each internal node
	 * and the rate).
	 *
	 * This constructor uses the standard definition of a rate (default value = 0, min value = 0, max value = 1)
	 * and does include the rate as the parameter
	 * <p>
	 * <em>This parameterisation of a clock-tree, ensuring that
	 * all parameters are independent of each other is due to
	 * Andrew Rambaut (personal communication).</em>
	 */
	public DatedTipsClockTree(Tree t, TimeOrderCharacterData tocd, boolean useDefaultParameters)
	{
		this(t,new ParameterizedDouble(0,0,1),tocd,(useDefaultParameters ? PARAMETERIZE_WITH_DEFAULTS : PARAMETERIZE_WITH_GIVEN) ,(useDefaultParameters ? PARAMETERIZE_WITH_DEFAULTS : PARAMETERIZE_WITH_GIVEN));
	}

	/**
	 * take any tree and afford it with an interface
	 * suitable for a clock-like tree with dated tips (parameters
	 * are the minimal node height differences at each internal node
	 * and the rate).
	 * <p>
	 * <em>This parameterisation of a clock-tree, ensuring that
	 * all parameters are independent of each other is due to
				 * Andrew Rambaut (personal communication).</em>
	 */
	public DatedTipsClockTree(Tree t, ParameterizedDouble rate, TimeOrderCharacterData tocd, int internalHeightMode, int rateMode)
	{
		this.rateMode_ = rateMode;
		this.internalHeightMode_ = internalHeightMode;

		this.rate = rate;
		setBaseTree(t);
		this.tocd = tocd;

		if (getRoot().getChildCount() < 2)
		{
			throw new IllegalArgumentException(
			"The root node must have at least two childs!");
		}

		switch(internalHeightMode) {
			case PARAMETERIZE_WITH_DEFAULTS : {
				parameter = new double[getInternalNodeCount()];
				for (int i = 0; i < parameter.length; i++) {
					parameter[i] = getDefaultValue(i);
				}
				break;
			}
			case PARAMETERIZE_WITH_GIVEN : {
				parameter = new double[getInternalNodeCount()];
				// assumes that the tree heights are already valid
				// DatedTipClockTree heights.
				// extracts rate on this assumption with aid of
				// TimeOrderCharacterData.
				heights2parameters();
				break;
			}
			case NO_PARAMETRIZE : {
				parameter = new double[0];
				break;
			}
		}
		date = new double[t.getExternalNodeCount()];
		getDates();

		if(rateMode==PARAMETERIZE_WITH_DEFAULTS) {
			boolean found = false;
			for(int i = 0 ;i < getExternalNodeCount() ; i++) {
				String name = getExternalNode(i).getIdentifier().getName();
				int index = tocd.whichIdNumber(name);
				double time = tocd.getTime(index);
				if (time > 0.0) {
					double height = getExternalNode(i).getNodeHeight();
					rate.setValue(height / time);
					System.out.println("rate = " + rate);
					found = true;
					break;
				}
			}
			if (!found) {
				rate.setValue(0.0);
			}
		}
		//If there is not direct parameterization we still alow external parameterization
		if(rateMode==NO_PARAMETRIZE) {
			rate.addExternalParameterListener(this);
		}
		// init
		parameters2Heights();
		NodeUtils.heights2Lengths(getRoot());
	}

	/** The rate was changed externally */
	public void parameterChanged(ParameterEvent pe) {
		parameters2Heights();
		NodeUtils.heights2Lengths(getRoot());
	}

	/** make parameters consistent with branch lengths and rate parameter */
	public void update()
	{
		//createNodeList();
		getDates(); // order of tips may have changed!
		NodeUtils.lengths2Heights(getRoot());
		heights2parameters();
	}

	// interface Parameterized (Rambaut paremeterisation)

	public int getNumParameters()
	{
		return
			((internalHeightMode_==NO_PARAMETRIZE) ? 0 : getInternalNodeCount() )+
			((rateMode_== NO_PARAMETRIZE)? 0 : 1);
	}

	private final boolean isHeightParameter(int n) {
		return ((internalHeightMode_!=NO_PARAMETRIZE)&& n < getInternalNodeCount());
	}
	public void setParameter(double param, int n)
	{
		if (isHeightParameter(n))
		{
			parameter[n] = param;
		}
		else
		{
			if(rateMode_!=NO_PARAMETRIZE) {
				rate.setValue(param);
			} else {
				throw new RuntimeException("Assertion error : Illegal attempt to change rate parameter");
			}
		}
		parameters2Heights();
		NodeUtils.heights2Lengths(getRoot());

	}

	public double getParameter(int n)
	{
		if (isHeightParameter(n))
		{
			return parameter[n];
		}
		else
		{
			return rate.getValue();
		}
	}

	public void setParameterSE(double paramSE, int n)
	{
		// we are only interested in SE of rate
		if (isHeightParameter(n))
		{
			// do nothing
		}
		else
		{
			rate.setSE(paramSE);
		}
		return;
	}

	/** return standard error of parameter */
	public double getParameterSE(int n)
	{
		// we are only interested in SE of rate
		if (isHeightParameter(n))
		{
			return 0.0;
		}
		else
		{
			return rate.getSE();
		}
	}



	public double getLowerLimit(int n)
	{
		if (isHeightParameter(n))
		{
			return BranchLimits.MINARC;
		}
		else
		{
			return rate.getLowerLimit();
		}
	}

	public double getUpperLimit(int n)
	{
		if (isHeightParameter(n))
		{
			return BranchLimits.MAXARC;
		}
		else
		{
			return rate.getUpperLimit(); // subst. per unit time
		}
	}

	public double getDefaultValue(int n)
	{
		if (isHeightParameter(n))
		{
			return BranchLimits.DEFAULT_LENGTH;
		}
		else
		{
			return rate.getDefaultValue(); // contemporaneous tips
		}
	}

	/** For other people to check and comment on - why the abs() stuff?
	 *	- Matthew
	 * Also - this method can be written considerably efficiently (some kindof
	 * tree walk might be better (can be handled by the tree class)
	 */
	private void updateRateStuff() {
		//System.out.println("Update rate stuff");
		//System.out.println("Max Date:"+maxDate);
		double rateValue = rate.getValue();
		int numberOfExternalNodes = getExternalNodeCount();
		for (int i = 0; i < numberOfExternalNodes; i++)
		{
			Node leaf = getExternalNode(i);
			double h = (maxDate-date[i])*rateValue;
			System.out.println("H:"+h);
			leaf.setNodeHeight(Math.abs(h));
			double hp = leaf.getParent().getNodeHeight();
			//System.out.println("Parent Height:"+leaf.getParent().getNodeHeight());
			leaf.setBranchLength(Math.abs(hp-h));
			//System.out.println("L:"+(hp-h));
		}
	}

	/**
	 * set rate (and thus node heights and branch lengths of leaves) without
	 * changing all other node heights and branch lengths
	 */
	public void setRate(double r)
	{
		rate.setValue(r);
		//If not parameterize will let call back method handle updating...
		if(rateMode_!=NO_PARAMETRIZE) {
			updateRateStuff();
		}
	}

	/**
	 * get rate
	 */
	public double getRate()
	{
		return rate.getValue();
	}


	/**
	 * set rate SE
	 */
	public void setRateSE(double rSE)
	{
		rate.setSE(rSE);
	}


	/**
	 * find max. rate (for setRate) allowed by current node heights
	 */
	public double getMaxRate()
	{
		double maxRate = 0;

		for (int i = 0; i < getExternalNodeCount(); i++)
		{
			Node leaf = getExternalNode(i);

			// maximum rate for this leaf
			double maxRateLeaf = (leaf.getParent().getNodeHeight()-BranchLimits.MINARC)/(maxDate-date[i]);

			if (i == 0 || maxRateLeaf < maxRate)
			{
				maxRate = maxRateLeaf;
			}
		}

		return maxRate;
	}

	/**
	 * Gets the TimeOrderCharacterData
	 */
	public TimeOrderCharacterData getTimeOrderCharacterData() {
		if (tocd != null) {
			return tocd;
		} else {
			//extract from tree.
			//WHAT UNITS TO USE!? Defaulting to EXPECTED_SUBSTITUTIONS

			TimeOrderCharacterData tempTOCD =
				new TimeOrderCharacterData(TreeUtils.getLeafIdGroup(this),
					Units.EXPECTED_SUBSTITUTIONS);
			tempTOCD.setTimes(date, Units.EXPECTED_SUBSTITUTIONS, true);

			return tempTOCD;
		}
	}

	//
	// Private stuff
	//

	private void parameters2Heights()	{
		// nodes have been stored by a post-order traversal
		for (int i = 0; i < getExternalNodeCount(); i++) {
			getExternalNode(i).setNodeHeight(  (maxDate-date[i])*rate.getValue()  );
		}
		if(internalHeightMode_!=NO_PARAMETRIZE) {
			for (int i = 0; i < getInternalNodeCount(); i++) {
				Node node = getInternalNode(i);
				node.setNodeHeight(parameter[i] + NodeUtils.findLargestChild(node));
			}
		}
	}

	private void heights2parameters()
	{
		for (int i = 0; i < getInternalNodeCount(); i++)
		{
			Node inode = getInternalNode(i);
			parameter[i] = inode.getNodeHeight()-NodeUtils.findLargestChild(inode);
		}
	}

	private void getDates()
	{

		double maxTime = 0.0;
		if (tocd != null) {
			for (int i = 0; i < date.length; i++) {
				if (tocd.getTime(i) > maxTime) {
					maxTime = tocd.getTime(i);
				}
			}
		}

		for (int i = 0; i < getExternalNodeCount(); i++)
		{
			if (tocd == null) {
				// extract dates from sequence identifiers
				date[i] = extractDate(getExternalNode(i).getIdentifier().getName());
			} else {

				if (tocd.hasTimes()) {
					String name = getExternalNode(i).getIdentifier().getName();

					int index = tocd.whichIdNumber(name);

					//times are measured backwards from zero at most recent tip.
					// this needs to be converted to forward-time dates starting
					// from time zero at oldest tip.

					date[i] = maxTime - tocd.getTime(index);
				} else {
					//no times available
					throw new IllegalArgumentException("TimeOrderCharacterData does not have any times!");
				}
			}
		}

		// find maximum and minimum date
		minDate = maxDate = date[0];
		for (int i = 1; i < getExternalNodeCount(); i++)
		{
			if (date[i] > maxDate) maxDate = date[i];
			if (date[i] < minDate) minDate = date[i];
		}

		// check for equality of dates
		if (minDate == maxDate)
		{
			throw new IllegalArgumentException("Tip dates must not be the same for all tips");
		}
	}

	private double extractDate(String string)
	{
		StringBuffer buffer = new StringBuffer();

		int len = string.length();

		boolean readDot = false;
		for (int i = len-1; i > -1; i--)
		{
			char c = string.charAt(i);

			if ( !Character.isDigit(c) &&  !(c == '.' && !readDot) ) break;
			buffer.append(c);
			if (c == '.') readDot = true;

		}
		buffer.reverse();

		String date = buffer.toString();

		if (date.length() == 0)
		{
			return 0.0;
		}
		else
		{
			return (Double.valueOf(buffer.toString())).doubleValue();
		}
	}

	private double[] parameter;
	private ParameterizedDouble rate;
	private double minDate, maxDate;
	private double[] date;
	private TimeOrderCharacterData tocd;
	private int internalHeightMode_;
	private int rateMode_;

}
